"""This folder contains message format for captian proto serialization.

Note: Each capnp message format should have unique hex identifier
(ex: @0xcd0709e35fffa8d8)
These can be generated in terminal by the command `capnp id` after pycapnp installation.
"""
