body[data-jp-theme-light="false"],
body.vscode-dark {
  --primary-color: #111111;
  --secondary-color: #212121;
  --tertiary-color: #cfcdd6;
  --button-color: #111111;
  --colors-black: #ffffff;
  --surface-color: #fff;
  --text-color: #ffffff;
}

body {
  --primary-color: #ffffff;
  --secondary-color: #f5f5f5;
  --tertiary-color: #000000de;
  --button-color: #d1d5db;
  --colors-black: #17161d;
  --surface-color: #464158;
  --text-color: #2e2b3b;
}

.header-1 {
  font-style: normal;
  font-weight: 600;
  font-size: 2.0736em;
  line-height: 100%;
  leading-trim: both;
  text-edge: cap;
  color: var(--colors-black);
}

.header-2 {
  font-style: normal;
  font-weight: 600;
  font-size: 1.728em;
  line-height: 100%;
  leading-trim: both;
  text-edge: cap;
  color: var(--colors-black);
}

.header-3 {
  font-style: normal;
  font-weight: 600;
  font-size: 1.44em;
  line-height: 100%;
  leading-trim: both;
  text-edge: cap;
  color: var(--tertiary-color);
}

.header-4 {
  font-style: normal;
  font-weight: 600;
  font-size: 1.2em;
  line-height: 100%;
  leading-trim: both;
  text-edge: cap;
  color: var(--colors-black);
}

.paragraph {
  font-style: normal;
  font-weight: 400;
  font-size: 14px;
  line-height: 100%;
  leading-trim: both;
  text-edge: cap;
  color: var(--text-color);
}

.paragraph-sm {
  font-family: "Roboto";
  font-style: normal;
  font-weight: 400;
  font-size: 11.62px;
  line-height: 100%;
  leading-trim: both;
  text-edge: cap;
  color: var(--text-color);
}

.code-text {
  font-family: "Consolas";
  font-style: normal;
  font-weight: 400;
  font-size: 13px;
  line-height: 130%;
  leading-trim: both;
  text-edge: cap;
  color: var(--text-color);
}

.numbering-entry {
  display: none;
}

/* Tooltip container */
.tooltip {
  position: relative;
  display: inline-block;
  border-bottom: 1px dotted black;
  /* If you want dots under the hoverable text */
}

/* Tooltip text */
.tooltip .tooltiptext {
  visibility: hidden;
  width: 120px;
  background-color: black;
  color: #fff;
  text-align: center;
  padding: 5px 0;
  border-radius: 6px;

  /* Position the tooltip text - see examples below! */
  position: absolute;
  z-index: 1;
}

.repr-cell {
  padding-top: 20px;
}

.text-bold {
  font-weight: bold;
}

.pr-8 {
  padding-right: 8px;
}

.pt-8 {
  padding-top: 8px;
}

.pl-8 {
  padding-left: 8px;
}

.pb-8 {
  padding-bottom: 8px;
}

.py-25 {
  padding-top: 25px;
  padding-bottom: 25px;
}

.flex {
  display: flex;
}

.gap-10 {
  gap: 10px;
}

.items-center {
  align-items: center;
}

.folder-icon {
  color: var(--tertiary-color);
}

.search-input {
  display: flex;
  flex-direction: row;
  align-items: center;
  padding: 8px 12px;
  width: 343px;
  height: 24px;
  /* Lt On Surface/Low */
  background-color: var(--secondary-color);
  border-radius: 30px;

  /* Lt On Surface/Highest */
  color: var(--tertiary-color);
  border: none;
  /* Inside auto layout */
  flex: none;
  order: 0;
  flex-grow: 0;
}

.search-input:focus {
  outline: none;
}

.search-input:focus::placeholder,
.search-input::placeholder {
  /* Chrome, Firefox, Opera, Safari 10.1+ */
  color: var(--tertiary-color);
  opacity: 1;
  /* Firefox */
}

.search-button {
  /* Search */
  leading-trim: both;
  text-edge: cap;
  display: flex;
  align-items: center;
  text-align: center;

  /* Primary/On Light */
  background-color: var(--button-color);
  color: var(--tertiary-color);

  border-radius: 30px;
  border-color: var(--secondary-color);
  border-style: solid;
  box-shadow:
    rgba(60, 64, 67, 0.3) 0px 1px 2px 0px,
    rgba(60, 64, 67, 0.15) 0px 1px 3px 1px;
  cursor: pointer;
  /* Inside auto layout */
  flex: none;
  order: 1;
  flex-grow: 0;
}

.grid-index-cells {
  grid-column: span 1;
  /* tmp fix to make left col stand out (fix with font-family) */
  font-weight: 600;
  background-color: var(--secondary-color) !important;
  color: var(--tertiary-color);
}

.center-content-cell {
  margin: auto;
}

.grid-header {
  /* Auto layout */
  display: flex;
  flex-direction: column;
  align-items: center;
  padding: 6px 4px;

  resize: horizontal;
  /* Lt On Surface/Surface */
  /* Lt On Surface/High */
  border: 1px solid #cfcdd6;
  /* tmp fix to make header stand out (fix with font-family) */
  font-weight: 600;
  background-color: var(--secondary-color);
  color: var(--tertiary-color);
}

.grid-row {
  display: flex;
  flex-direction: column;
  align-items: flex-start;
  padding: 6px 4px;
  overflow: hidden;
  border: 1px solid #cfcdd6;
  background-color: var(--primary-color);
  color: var(--tertiary-color);
}

.syncstate-col-footer {
  font-family: "DejaVu Sans Mono", "Open Sans";
  font-size: 12px;
  font-weight: 400;
  line-height: 16.8px;
  text-align: left;
  color: #5e5a72;
}

.syncstate-description {
  font-family: Open Sans;
  font-size: 14px;
  font-weight: 600;
  line-height: 19.6px;
  text-align: left;
  white-space: nowrap;
  flex-grow: 1;
}

.widget-header2 {
  display: flex;
  gap: 8px;
  justify-content: start;
  width: 100%;
  overflow: hidden;
  align-items: center;
}

.widget-header2-2 {
  display: flex;
  gap: 8px;
  justify-content: start;
  align-items: center;
}

.jobs-title {
  font-family:
    Open Sans,
    sans-serif;
  font-size: 18px;
  font-weight: 600;
  line-height: 25.2px;
  text-align: left;
  color: #1f567a;
}

.diff-state-orange-text {
  color: #b8520a;
}

.diff-state-no-obj {
  font-family: "DejaVu Sans Mono", "Open Sans";
  font-size: 12px;
  font-weight: 400;
  line-height: 16.8px;
  text-align: left;
  color: #5e5a72;
}

.diff-state-intro {
  font-family: Open Sans;
  font-size: 14px;
  font-weight: 400;
  line-height: 19.6px;
  text-align: left;
  color: #b4b0bf;
}

.diff-state-header {
  font-family: Open Sans;
  font-size: 22px;
  font-weight: 600;
  line-height: 30.8px;
  text-align: left;
  color: #353243;
  display: flex;
  gap: 8px;
}

.diff-state-sub-header {
  font-family: Open Sans;
  font-size: 14px;
  font-weight: 400;
  line-height: 19.6px;
  text-align: left;
  color: #5e5a72;
}

.badge {
  /* code-text; */
  border-radius: 30px;
  padding: 6px;
  white-space: nowrap;
  overflow: hidden;
  line-height: 1.2;
  font-family: monospace;
}

.label {
  /* code-text; */
  border-radius: 4px;
  padding: 6px 4px;
  white-space: nowrap;
  overflow: hidden;
  line-height: 1.2;
  font-family: monospace;
}

.label-light-purple {
  /* label; */
  background-color: #c9cfe8;
  color: #373b7b;
}

.label-light-blue {
  /* label; */
  background-color: #c2def0;
  color: #1f567a;
}

.label-orange {
  /* badge; */
  background-color: #fee9cd;
  color: #b8520a;
}

.label-gray {
  /* badge; */
  background-color: #ecebef;
  color: #353243;
}

.label-green {
  /* badge; */
  background-color: #d5f1d5;
  color: #256b24;
}

.label-red {
  /* label; */
  background-color: #f2d9de;
  color: #9b2737;
}

.badge-blue {
  /* badge; */
  background-color: #c2def0;
  color: #1f567a;
}

.badge-purple {
  /* badge; */
  background-color: #c9cfe8;
  color: #373b7b;
}

.badge-green {
  background-color: #d5f1d5;
  color: #256b24;
}

.badge-red {
  /* badge; */
  background-color: #f2d9de;
  color: #9b2737;
}

.badge-gray {
  /* badge; */
  background-color: #ecebef;
  color: #2e2b3b;
}

.paginationContainer {
  width: 100%;
  /*height: 30px;*/
  display: flex;
  justify-content: center;
  gap: 8px;
  padding: 5px;
  color: var(--tertiary-color);
}

.widget-label-basic {
  display: flex;
}

.widget-label-basic input[type="checkbox"][disabled] {
  filter: sepia(0.3) hue-rotate(67deg) saturate(3);
}

.page {
  color: black;
  font-weight: bold;
  color: var(--tertiary-color);
}

.page:hover {
  color: #38bdf8;
  cursor: pointer;
}

.clipboard:hover {
  cursor: pointer;
  color: var(--tertiary-color);
}

.rendered_html tbody tr:nth-child(odd) {
  background: transparent;
}

.search-field {
  display: flex;
  align-items: center;
  border-radius: 30px;
  background-color: var(--secondary-color);
}

.syft-dropdown {
  margin: 5px;
  margin-left: 5px;
  position: relative;
  display: inline-block;
  text-align: center;
  background-color: var(--button-color);
  min-width: 100px;
  padding: 2px;
  border-radius: 30px;
}

.syft-dropdown:hover {
  cursor: pointer;
}

.syft-dropdown-content {
  margin-top: 26px;
  display: none;
  position: absolute;
  min-width: 100px;
  box-shadow: 0px 8px 16px 0px rgba(0, 0, 0, 0.2);
  padding: 12px 6px;
  z-index: 1;
  background-color: var(--primary-color);
  color: var(--tertiary-color);
}

.dd-options {
  padding-top: 4px;
}

.dd-options:first-of-type {
  padding-top: 0px;
}

.dd-options:hover {
  cursor: pointer;
  background: #d1d5db;
}

.arrow {
  border: solid black;
  border-width: 0 3px 3px 0;
  display: inline-block;
  padding: 3px;
}

.down {
  transform: rotate(45deg);
  -webkit-transform: rotate(45deg);
}

.syft-widget ul {
  list-style-type: none;
  margin: 0;
  padding: 0;
  overflow: hidden;
}

.syft-widget li {
  float: left;
  border-bottom: solid;
  border-bottom-color: #cfcdd6;
}

.syft-widget li a {
  display: block;
  text-align: center;
  padding: 14px 16px;
  color: #cfcdd6;
}

.log-tab-header {
  border-bottom: solid 2px #ecebef;
  padding: 4px 16px;
}

.active-border {
  border-bottom: solid 2px #1f567a;
  font-weight: 700;
}

.active {
  color: #1f567a;
}

.syft-widget li a:hover {
  background-color: #c2def0;
}

.syft-user_code,
.syft-project,
.syft-project-create,
.syft-settings,
.syft-dataset,
.syft-asset,
.syft-contributor,
.syft-request,
.syft-syncstate,
.job-info {
  color: var(--surface-color);
}

.syft-dataset h1,
.syft-dataset h2,
.syft-dataset h3,
.syft-dataset p,
.syft-asset h3,
.syft-asset p,
.syft-syncstate h3,
.syft-syncstate p {
  font-family: "Open Sans";
}

.diff-container {
  border: 0.5px solid #b4b0bf;
}

.syft-container {
  padding: 5px;
  font-family: "Open Sans";
}

.syft-alert-info {
  color: #1f567a;
  background-color: #c2def0;
  border-radius: 4px;
  padding: 5px;
  padding: 13px 10px;
}

.syft-code-block {
  background-color: #f7f7f7;
  border: 1px solid #cfcfcf;
  padding: 0px 2px;
}

.syft-space {
  margin-top: 1em;
}
